﻿--[[
Name: Rating Buster
Revision: $Revision: 2 $
Developed by: Whitetooth@Cenarius (hotdogee@巴哈姆特baha)
Description: Converts combat ratings in tooltips into normal percentages.
]]

-----------------
-- Addon Setup --
-----------------
-- Version Info
local MAJOR_VERSION = "0.7.5"
--local MINOR_VERSION = tonumber((string.gsub("$Revision: 2 $", "^.-(%d+).-$", "%1")))
-- Librarys
local Compost = AceLibrary("Compost-2.0")
local L = AceLibrary("AceLocale-2.2"):new("RatingBuster")
-- AceAddon Initialization
RatingBuster = AceLibrary("AceAddon-2.0"):new("AceDB-2.0", "AceConsole-2.0", "AceEvent-2.0", "AceDebug-2.0")
RatingBuster.title = "Rating Buster"
RatingBuster.version = MAJOR_VERSION
--RatingBuster.date = string.gsub("$Date: 2006-11-05 15:10:31 -0700 (Sun, 03 Sep 2006) $", "^.-(%d%d%d%d%-%d%d%-%d%d).-$", "%1")

-- Set Debugging
RatingBuster:SetDebugging(false)


---------------------
-- Saved Variables --
---------------------
-- Register DB
RatingBuster:RegisterDB("RatingBusterDB")
-- Default values
RatingBuster.DBDefaults = {
	showItemLevel = true,
	showItemID = false,
	useRequiredLevel = true,
}
-- Register Defaults
RatingBuster:RegisterDefaults("profile", RatingBuster.DBDefaults)


---------------------------
-- Slash Command Options --
---------------------------
local consoleOptions = {
	type='group',
	args = {
		itemid = {
			type = 'toggle',
			name = L["Show Item ID"],
			desc = L["Toggle display of Item ID in tooltip"],
			get = function() return RatingBuster.db.profile.showItemID end,
			set = function(v) RatingBuster.db.profile.showItemID = v end,
		},
		itemlevel = {
			type = 'toggle',
			name = L["Show Item Level"],
			desc = L["Toggle display of Item Level in tooltip"],
			get = function() return RatingBuster.db.profile.showItemLevel end,
			set = function(v) RatingBuster.db.profile.showItemLevel = v end,
		},
		usereqlv = {
			type = 'toggle',
			name = L["Use Required Level"],
			desc = L["Toggle calculate using the required level if you are below the required level"],
			get = function() return RatingBuster.db.profile.useRequiredLevel end,
			set = function(v) RatingBuster.db.profile.useRequiredLevel = v end,
		},
		targetlevel = {
			type = 'range',
			name = L["Target Level"],
			desc = L["Changes the Target Level used in rating calculations"],
			get = function() return RatingBuster.targetLevel end,
			set = function(v) RatingBuster.targetLevel = v end,
			min = 10,
			max = 73, -- set to level cap + 3
			step = 1,
		},
	},
}

RatingBuster:RegisterChatCommand({"/rb", "/ratingbuster"}, consoleOptions)


----------------------------
-- ReverseRating Function --
----------------------------
-- Rating ID
RB_WEAPON_SKILL = 1
RB_DEFENSE_SKILL = 2
RB_DODGE = 3
RB_PARRY = 4
RB_BLOCK = 5
RB_HIT_MELEE = 6
RB_HIT_RANGED = 7
RB_HIT_SPELL = 8
RB_CRIT_MELEE = 9
RB_CRIT_RANGED = 10
RB_CRIT_SPELL = 11
RB_HASTE_MELEE = 12
RB_HASTE_RANGED = 13
RB_HASTE_SPELL = 14
RB_RESILIENCE = 15
RB_HIT_TAKEN_MELEE = 16

-- Level 60 base rating
local RatingBase = {
	[RB_WEAPON_SKILL] = 2.5,
	[RB_DEFENSE_SKILL] = 1.5,
	[RB_DODGE] = 12,
	[RB_PARRY] = 20,
	[RB_BLOCK] = 5,
	[RB_HIT_MELEE] = 10,
	[RB_HIT_RANGED] = 10,
	[RB_HIT_SPELL] = 8,
	[RB_CRIT_MELEE] = 14,
	[RB_CRIT_RANGED] = 14,
	[RB_CRIT_SPELL] = 14,
	[RB_HASTE_MELEE] = 10,
	[RB_HASTE_RANGED] = 10,
	[RB_HASTE_SPELL] = 10,
	[RB_RESILIENCE] = 25,
	[RB_HIT_TAKEN_MELEE] = 10,
}

-- Formula reverse engineered by Whitetooth@Cenarius (hotdogee@巴哈姆特baha)
-- returns percentages
function RatingBuster.ReverseRating(rNumber, rType, level)
	-- check for invalid input
	if type(rNumber) ~= "number" or rType < 1 or rType > 16 then return end
	-- defaults to player level if not given
	level = level or UnitLevel("player")
	if level >= 60 then
		return rNumber/RatingBase[rType]*((-3/82)*level+(131/41))
	elseif level >= 10 then
		return rNumber/RatingBase[rType]/((1/52)*level-(8/52))
	else
		return 0
	end
end


---------------------
-- Initializations --
---------------------
function RatingBuster:OnInitialize()
	-- Build the RatingNameToID table from L["ratingNames"] table
	self:BuildRatingNameToID()
	-- Hooking moved from OnEnable to OnInitialize until AceHook is fixed, to prevent multi hooking when resuming from standby
	-- Hook tooltip SetX methods
	self:HookTooltipMethods()
	-- OnUpdate Frame
	self.OnUpdateFrame = CreateFrame("Frame")
	self.OnUpdateFrame:SetScript("OnUpdate", self.OnUpdate)
end

function RatingBuster:BuildRatingNameToID()
	RatingBuster.RatingNameToID = {}
	for _, v in ipairs(L["ratingNames"]) do
		self.RatingNameToID[v[1]] = v[2]
	end
end

function RatingBuster:OnEnable()
	-- Initialize targetLevel
	RatingBuster.targetLevel = UnitLevel("player")
	-- Check player level for lowbies
	if RatingBuster.targetLevel < 10 then
		RatingBuster.targetLevel = 10
	end
	-- for setting a new level
	self:RegisterEvent("PLAYER_LEVEL_UP")
end

function RatingBuster:OnDisable()
end

-- event = PLAYER_LEVEL_UP
-- arg1 = New player level
function RatingBuster:PLAYER_LEVEL_UP()
	RatingBuster.targetLevel = arg1
end


---------------
-- Schedules --
---------------
-- Schedule Table
RatingBuster.ScheduledTasks = {}
function RatingBuster:Schedule(taskName, timeAfter, functionName, ...)
	if (taskName and timeAfter) then -- functionName not required so we can use IsScheduled as a timer check
		Compost:Reclaim(self.ScheduledTasks[taskName], 2)
		self.ScheduledTasks[taskName] = Compost:AcquireHash(
			"TargetTime", GetTime() + timeAfter,
			"FunctionName", functionName,
			"Arg", Compost:Acquire(...)
		)
	end
end

function RatingBuster:ScheduleRepeat(taskName, repeatRate, functionName, ...)
	if (taskName and repeatRate and functionName) then -- functionName required
		Compost:Reclaim(self.ScheduledTasks[taskName], 2)
		self.ScheduledTasks[taskName] = Compost:AcquireHash(
			"Elapsed", 0,
			"RepeatRate", repeatRate,
			"FunctionName", functionName,
			"Arg", Compost:Acquire(...)
		)
	end
end

function RatingBuster:UnSchedule(taskName)
	--WT_RaidWarningAPI.Announce({HOTDOG = taskName})
	if not taskName then return end
	Compost:Reclaim(self.ScheduledTasks[taskName], 2)
	self.ScheduledTasks[taskName] = nil
end

function RatingBuster:IsScheduled(taskName)
	return (self.ScheduledTasks[taskName] ~= nil)
end

function RatingBuster:OnUpdate(elapsed)
	--RatingBuster:Debug("update: "..tostring(elapsed))
	local currentTime = GetTime()
	for taskName, task in pairs(RatingBuster.ScheduledTasks) do
		if type(task.TargetTime) ~= "nil" and (currentTime >= task.TargetTime) then
			RatingBuster:UnSchedule(taskName)
			if (task.FunctionName) then
				task.FunctionName(unpack(task.Arg))
			end
		elseif type(task.Elapsed) ~= "nil" then
			task.Elapsed = task.Elapsed + arg1
			if (task.Elapsed >= task.RepeatRate) then
				task.FunctionName(unpack(task.Arg))
				task.Elapsed = task.Elapsed - task.RepeatRate
			end
		end
	end
end


-------------------------------
-- Hook Tooltip SetX Methods --
-------------------------------
RatingBuster.TooltipList = {
	ItemRefTooltip,
	GameTooltip,
	ShoppingTooltip1,
	ShoppingTooltip2,
	--EquipCompare support
	ComparisonTooltip1,
	ComparisonTooltip2,
	--EQCompare support
	EQCompareTooltip1,
	EQCompareTooltip2,
	-- socketing interface
	--ItemSocketingDescription, -- commented out cause can't get it to work
	-- LinkWrangler support
	IRR_ItemRefTooltip1,
	IRR_ItemCompTooltip1,
	IRR_ItemCompTool11,
	IRR_ItemRefTooltip2,
	IRR_ItemCompTooltip2,
	IRR_ItemCompTool12,
	IRR_ItemRefTooltip3,
	IRR_ItemCompTooltip3,
	IRR_ItemCompTool13,
	IRR_ItemRefTooltip4,
	IRR_ItemCompTooltip4,
	IRR_ItemCompTool14,
	IRR_ItemRefTooltip5,
	IRR_ItemCompTooltip5,
	IRR_ItemCompTool15,
	-- MultiTips support
	ItemRefTooltip2,
	ItemRefTooltip3,
	ItemRefTooltip4,
	ItemRefTooltip5,
}

RatingBuster.MethodList = {
	"SetHyperlink",
	"SetBagItem",
	"SetInventoryItem",
	-- auction
	"SetAuctionItem",
	"SetAuctionSellItem",
	"SetAuctionCompareItem",
	-- loot
	"SetLootItem",
	"SetLootRollItem",
	-- crafting
	"SetCraftSpell",
	"SetCraftItem",
	"SetTradeSkillItem",
	"SetTrainerService",
	-- mail
	"SetInboxItem",
	"SetSendMailItem",
	-- buffs
	"SetPlayerBuff",
	-- quest log
	"SetQuestItem",
	"SetQuestLogItem",
	-- trade
	"SetTradePlayerItem",
	"SetTradeTargetItem",
	-- vendor tooltip
	"SetMerchantItem",
	"SetMerchantCompareItem",
	"SetBuybackItem",
	-- socketing interface
	--"SetSocketedItem", -- main display
	"SetSocketGem",
	"SetExistingSocketGem",
}

function RatingBuster.SetItem(tooltip)
	if not tooltip.GetItem then return end
	local _, link = tooltip:GetItem()
	RatingBuster.ProcessTooltip(tooltip, link)
end

function RatingBuster:HookTooltipMethods()
	for _, tooltip in ipairs(self.TooltipList) do
		if tooltip then
			for _, methodName in ipairs(self.MethodList) do
				hooksecurefunc(tooltip, methodName, RatingBuster.SetItem)
				-- prints out an error when put on standby, as of Rev: 15813
				--self:SecureHook(tooltip, methodName, self.ProcessTooltip)
			end
		end
	end
end

--------------------------
-- Process Tooltip Core --
--------------------------
--[[
Method 1
"+15 Crit Rating"
-> "+15(1.20%) Crit Rating"
"Equip: Increases your hit rating by 10."
-> "Equip: Increases your hit rating by 10(1.20%)."
Method 2
"+15 Crit Rating"
-> "+15 Crit Rating (1.20%)"
"Equip: Increases your hit rating by 10."
-> "Equip: Increases your hit rating by 10. (1.20%)"
Method 3
"+15 Crit Rating"
-> "+15 Crit Rating (1.20%)"
"Equip: Increases your hit rating by 10."
-> "Equip: +10 Hit Rating (1.20%)"
Method 4
+10 Sta (100HP) 
Equip: Adds 15 critical hit rating (.85% ) 
Equip: Adds 12 resilience (-.58% crit, -1.16% crit damage) 
--]]
function RatingBuster.ProcessTooltip(tooltip, link)
	-- check if in standby mode
	if not RatingBuster:IsActive() then return end
	-- Get Required Level
	local requiredLevel
	if RatingBuster.db.profile.useRequiredLevel and link then
		_, _, _, _, requiredLevel = GetItemInfo(link)
		--RatingBuster:Print(link..", "..requiredLevel)
		if requiredLevel and requiredLevel <= RatingBuster.targetLevel then
			requiredLevel = nil
		end
	end
	--Anlegen: Erhöht die Chance aller Gruppenmitglieder, die sich im Umkreis von 30 befinden, auf einen kritischen Treffer mit Zaubern um 28%. 
	for i = 2, tooltip:NumLines() do
		local fontString = getglobal(tooltip:GetName().."TextLeft"..i)
		local lineText = fontString:GetText()
		if lineText and strfind(lineText, "%d") then -- do nothing if we don't find a number
			for _, p in ipairs(L["numberPatterns"]) do
				local text = lineText
				local lowerText = string.lower(text) -- convert to lower so we don't have to worry about same ratings with different cases
				-- Capture the increased amount
				local s, e, amount = strfind(lowerText, p.pattern)
				if amount then
					-- Check for separators
					local separator
					for _, s in ipairs(p.separators) do
						if strfind(lineText, s) then
							separator = s
							-- Replace separator strings found in lineText with a single character "/", so we can use strsplit()
							lineText = string.gsub(lineText, s, "/")
							-- only need support for 1 kind of separator per line, we will use the first one found, so break out here
							break
						end
					end
					-- If this line has a separator, then we do the split and combine algorithm
					if separator then
						local splitText = Compost:Acquire(strsplit("/", lineText))
						--RatingBuster:Debug(p.pattern)
						for i, text in ipairs(splitText) do
							--RatingBuster:Debug(text)
							local lowerText = string.lower(text) -- convert to lower so we don't have to worry about same ratings with different cases
							-- Capture the increased amount
							local s, e, amount = strfind(lowerText, p.pattern)
							if amount then
								-- Capture the rating name
								local ratingID
								for _, v in ipairs(L["ratingNames"]) do
									if strfind(lowerText, v[1]) then
										ratingID = RatingBuster.RatingNameToID[v[1]]
										break
									end
								end
								if ratingID then
									-- calculate reverse value
									local reversedAmount = RatingBuster.ReverseRating(tonumber(amount), ratingID, requiredLevel or RatingBuster.targetLevel)
									--RatingBuster:Debug(reversedAmount..", "..amount..", "..ratingID..", "..RatingBuster.targetLevel)-- debug
									-- build reversed string
									local reversedString
									-- check if rating is a skill type
									if ratingID == 1 or ratingID == 2 then
										reversedString = "("..format("%.2f", reversedAmount)..")"
									-- check if rating is resilience
									elseif ratingID == 15 then
										reversedString = "(-"..format("%.2f", reversedAmount).."%%%%)"
									else
										reversedString = "("..format("%.2f", reversedAmount).."%%%%)"
									end
									-- build converted string
									splitText[i] = string.gsub(text, p.pattern, string.gsub(strsub(text, s, e), "%d+", "%0 "..reversedString), 1)
								end
							end
						end -- for i, text in ipairs(splitText)
						lineText = strjoin(separator, unpack(splitText))
						Compost:Reclaim(splitText)
					else
						-- Capture the rating name
						local ratingID
						for _, v in ipairs(L["ratingNames"]) do
							if strfind(lowerText, v[1]) then
								ratingID = RatingBuster.RatingNameToID[v[1]]
								break
							end
						end
						if ratingID then
							-- calculate reverse value
							local reversedAmount = RatingBuster.ReverseRating(tonumber(amount), ratingID, requiredLevel or RatingBuster.targetLevel)
							--RatingBuster:Debug(reversedAmount..", "..amount..", "..ratingID..", "..RatingBuster.targetLevel)-- debug
							-- build reversed string
							local reversedString
							-- check if rating is a skill type
							if ratingID == 1 or ratingID == 2 then
								reversedString = "("..format("%.2f", reversedAmount)..")"
							-- check if rating is resilience
							elseif ratingID == 15 then
								reversedString = "(-"..format("%.2f", reversedAmount).."%%%%)"
							else
								reversedString = "("..format("%.2f", reversedAmount).."%%%%)"
							end
							-- build converted string
							lineText = string.gsub(text, p.pattern, string.gsub(strsub(text, s, e), "%d+", "%0 "..reversedString), 1)
						end
					end
				end
			end
			-- SetText
			fontString:SetText(lineText)
		end -- if text then
	end -- for i = 2, tooltip:NumLines() do
	-- Item Level and Item ID
	-- Check for ItemLevel addon, do nothing if found
	if not ItemLevel_AddInfo and (RatingBuster.db.profile.showItemLevel or RatingBuster.db.profile.showItemID)then
		local id, level
		-- Get the Item ID from the link string
		if link then
			_, link, _, level = GetItemInfo(link)
			if link then
				_, _, id = strfind(link, "item:(%d+)")
			end
			local newLine = ""
			if level and RatingBuster.db.profile.showItemLevel then
				newLine = newLine..L["ItemLevel: "]..level
			end
			if id and RatingBuster.db.profile.showItemID then
				if newLine ~= "" then
					newLine = newLine..", "
				end
				newLine = newLine..L["ItemID: "]..id
			end
			if newLine ~= "" then
				tooltip:AddLine(newLine)
			end
		end
	end
	tooltip:Show()
end
